import collections
import copy
import datetime
import enum
import functools
import inspect
import logging


class Stopwatch:

    def __init__(
        self,
        stop_message=None,
        writer=print,
        now_function=datetime.datetime.utcnow
    ):
        self.__now = now_function
        self.__is_running = False
        self.__start_time = None
        self.__lap_start_time = None
        self.__stop_message = stop_message
        self.__writer = writer

    @property
    def is_running(self):
        return self.__is_running

    def start(self):
        assert not self.__is_running
        self.__start_time = self.now()
        self.__lap_start_time = self.__start_time
        self.__is_running = True

    def stop(self):
        assert self.__is_running
        self.__stop_time = self.now()
        self.__is_running = False

    def restart(self):
        self.stop()
        self.start()

    def elapsed_time(self, return_lap_time=False):
        elapsed_time_result = None
        lap_time_result = None
        if self.__start_time is None:
            elapsed_time_result = None
            lap_time_result = None
        elif self.__is_running:
            now = self.now()
            elapsed_time_result = now - self.__start_time
            lap_time_result = now - self.__lap_start_time
            if return_lap_time:
                self.__lap_start_time = now
        else:
            elapsed_time_result = self.__stop_time - self.__start_time
            lap_time_result = self.__stop_time - self.__lap_start_time

        if return_lap_time:
            return (elapsed_time_result, lap_time_result)
        else:
            return elapsed_time_result

    def now(self):
        return self.__now()

    # Context manager methods

    def __enter__(self):
        self.start()

    def __exit__(self, type, value, traceback):
        self.stop()
        if self.__stop_message and not traceback:
            self.__writer(self.__stop_message.format(
                elapsed_time=self.elapsed_time()
            ))
        if traceback:
            return False
        return True


class Variable:

    def __init__(self, *, value=None, name=None):
        self.name = name
        self._value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    def __call__(self):
        pass

    def __str__(self):
        if self.name is None:
            return repr(self)
        return self.name


def print_value(input_cell, message):
    def f():
        print(message.format(value=input_cell.value))
    return f


def arg_getter(arg):

    # If the arg is an object with a "value" attribute, we return a function
    # that will return the value when called.
    if hasattr(arg, "value"):
        def value_getter():
            return arg.value
        return value_getter

    # If the arg is a sequence, we first generate a list of getters for
    # the elements of the list. When then return a function that invokes
    # those getters when it is invoked.
    if isinstance(arg, (list, tuple)):
        # Here's the list of getters for the sequence elements
        getters = [arg_getter(x) for x in arg]

        # Here's the function that will invoke those getters when it is
        # invoked
        def sequence_getter():
            return type(arg)(g() for g in getters)
        return sequence_getter

    if not isinstance(arg, str) and \
       isinstance(arg, collections.abc.Collection):
        raise Exception(f"Don't know what to do with arg of type {type(arg)}")

    # Arg is neither an object with a "value" attribute and nor is it a
    # collection or a map. In this case, we consider the value to be a literal.
    # We generate a function that simply returns the value as is
    def literal_getter():
        return arg
    return literal_getter


def assign(destination, function, *arguments):
    """Call a function with given argument nodes and assign the result
    to a destination variable.

    Args:
        - destination   - a node implementing the 'Variable' interface
        - function      - a callable or a coroutine function
        - arguments     - nodes to supply the arguments to the function

    Returns:
        - depending on the type of the 'function' argument, a callable or
          a coroutine function
    """
    getters = [arg_getter(argument) for argument in arguments]

    if inspect.iscoroutinefunction(function):
        async def assigner():
            args = [getter() for getter in getters]
            destination.value = await function(*args)
        return assigner

    else:
        def assigner():
            args = [getter() for getter in getters]
            destination.value = function(*args)
        return assigner


class GraphException(Exception):

    def __init__(self, message):
        super(GraphException, self).__init__(message)


class CircularDependency(GraphException):

    def __init__(self, message="Circular dependency detected in graph"):
        super(CircularDependency, self).__init__(message)

# The following function implements Tarjan's algorthim for topological sort,
# which I got from the web. I don't know where exactly.
#
# Algorithm:
#
# L = Empty list that will contain the sorted nodes
# while there are unmarked nodes do
#   select an unmarked node n
#   visit(n)
#
# function visit(node n)
#    if n has a temporary mark then stop (not a DAG)
#    if n is not marked (i.e. has not been visited yet) then
#       mark n temporarily
#       for each node m with an edge from n to m do
#           visit(m)
#       unmark n temporarily
#       mark n permanently
#       add n to head of L
#


def topological_sort(dependency_dict: dict):
    """ 'dependency_dict' is a dict describing dependencies between nodes.
    There must be a key for each node in the graph and the value is a list of
    nodes on which the key is dependent. If a node is present in a dependency
    list but has not dependency list of its own, it is assumed to have no
    dependencies.

    Args:
        . dependency_dict
    """

    # Extract list of nodes from dependency_dict
    nodes = set(dependency_dict.keys()).union(*dependency_dict.values())

    ordering = []

    class Mark(enum.Enum):

        NONE = 1
        TEMPORARY = 2
        MARKED = 3

    marks = {node: Mark.NONE for node in nodes}

    def visit_node(node):
        mark = marks[node]
        if mark == Mark.TEMPORARY:
            raise CircularDependency()
        assert mark == Mark.NONE or mark == Mark.MARKED
        if mark == Mark.MARKED:
            return
        assert mark == Mark.NONE
        marks[node] = Mark.TEMPORARY
        for dependency in dependency_dict.get(node, []):
            visit_node(dependency)

        marks[node] = Mark.MARKED
        ordering.append(node)

    for node in nodes:
        visit_node(node)

    return ordering


class DependencyGraph:

    def __init__(self, initial_dependencies={}):
        self._topological_ordering = None
        self._precedence_dict = None
        self._input_nodes = None
        self.dependency_dict = {}
        for node, dependencies in initial_dependencies.items():
            self.add_node(node)
            for dependency in dependencies:
                self.add_dependency(node, dependency)

    def __len__(self):
        return len(self.dependency_dict)

    def __str__(self):
        return str(self.dependency_dict)

    def _dirty(self):
        self._topological_ordering = None
        self._precedence_dict = None
        self._input_nodes = None

    def add_dependency(self, from_cell, to_cell):
        """Declare a dependency between two nodes. The nodes are implicitly
        added to the graph if they are not yet in it.
        """
        if from_cell is None:
            raise ValueError(f"Invalid argument: from_cell={from_cell}")
        if to_cell is None:
            raise ValueError(f"Invalid argument: to_cell={to_cell}")
        self.add_node(from_cell)
        self.add_node(to_cell)
        if to_cell in self.dependency_dict[from_cell]:
            return
        self._dirty()
        self.dependency_dict[from_cell].add(to_cell)

    def add_precedence(self, from_cell, to_cell):
        self.add_dependency(to_cell, from_cell)

    def add_node(self, node):
        """This function adds a node to the graph without specifying a
        dependency.
        """
        if node is None:
            raise ValueError(f"Invalid argument: node={node}")
        if node in self.dependency_dict:
            return node
        self.dependency_dict[node] = set()
        self._dirty()
        return node

    @property
    def topological_ordering(self):
        if self._topological_ordering is None:
            self._topological_ordering = topological_sort(self.dependency_dict)
        return self._topological_ordering

    def check_for_cycles(self):
        # This has the side-effect of checking for cycles. Also, it is only
        # checked if the graph has been modified since the last time
        # topological_ordering was called.
        self.topological_ordering

    @property
    def precedence_dict(self):
        if self._precedence_dict is None:
            self._precedence_dict = {}
            for dependent, dependencies in self.dependency_dict.items():
                if dependent not in self._precedence_dict:
                    self._precedence_dict[dependent] = set()
                for dependency in dependencies:
                    if dependency not in self._precedence_dict:
                        self._precedence_dict[dependency] = set()
                    self._precedence_dict[dependency].add(dependent)
        return self._precedence_dict

    @property
    def has_cycle(self):
        try:
            self.topological_ordering
            return False
        except GraphException:
            return True

    @property
    def input_nodes(self):
        if self._input_nodes is None:
            self._input_nodes = {
                node for node, dependencies in self.dependency_dict.items()
                if len(dependencies) == 0
            }
        return self._input_nodes

    def decorate(self, decorator_factory):
        """This function replaces nodes in the graph with asyncio.Tasks.
        """
        decorators = {}

        def decorated(node):
            decorator = decorators.get(node, None)
            if decorator is None:
                decorator = decorator_factory(node)
                decorators[node] = decorator
            return decorator

        decorated_graph = DependencyGraph()
        for node in self.dependency_dict.keys():
            decorated_node = decorated_graph.add_node(decorated(node))
            for dependency in self.dependency_dict[node]:
                decorated_dependency = decorated(dependency)
                decorated_graph.add_dependency(decorated_node,
                                               decorated_dependency)
        return decorated_graph


def compute_sequential(graph):
    for node in graph.topological_ordering:
        node()


class TopologicalQueue:
    """TopologicalQueue is used by the `aio` package to implement concurrent
    graph computation.

    The method `ready_set` returns all nodes that have no incomplete
    dependencies.

    The method `completed` indicates that a node is complete. This will
    result in `ready_set` changing. The newly completed node will be removed
    and, potentially, some new node will become ready if that node was
    their last remaining incomplete dependency. For convenience, this method
    returns the `now ready set`, which is the set of nodes that were _not_
    ready before this method was called but have now _become_ ready as a
    result of this call.

    `empty` is used to test whether there are any remaining nodes in the
    queue.

    This class is based on functools.TopologicalSorter that will be included
    in Python 3.9. However, it is desiged to work specifically with
    `DependencyGraph` in this module rather than being general.
    """

    def __init__(self, dependency_graph):
        if dependency_graph.has_cycle:
            raise CircularDependency()

        self.dependency_graph = dependency_graph

        # To determine whether a node is read or not, we keep a dict that
        # contains the count of incomplete dependencies for each node. Once
        # that count is zero, the node is ready
        self.dependency_count = {
            node: len(dependencies)
            for node, dependencies in dependency_graph.dependency_dict.items()
        }

        # The ready set is the set of all nodes with no incomplete dependencies
        self.ready = {node for node, dependency_count in
                      self.dependency_count.items() if dependency_count == 0}

        # Note that a node is left in self.dependency_count even if its
        # dependency_count is 0. It will only be removed when it is completed.

    def __len__(self):
        return len(self.dependency_count)

    def empty(self):
        return len(self.dependency_count) == 0

    def ready_set(self):
        return set(self.ready)

    def completed(self, node):
        # Remove node from ready set
        self.ready.remove(node)

        now_ready = set()
        # Remove node from each of its dependents
        dependents = self.dependency_graph.precedence_dict[node]
        for dependent in dependents:
            self.dependency_count[dependent] -= 1
            if self.dependency_count[dependent] == 0:
                # If the dependent now has no incomplete dependencies, it is
                # ready
                self.ready.add(dependent)
                now_ready.add(dependent)
                # As mentioned in the constructor, a node is left in
                # self.dependency_count even if its dependency_count is 0. It
                # will only be removed when it is completed.

        # The node is complete and is removed from dependency_count.
        # When dependency_count is empty, the queue is also empty.
        del self.dependency_count[node]

        return now_ready



#!/usr/bin/env python3

import asyncio
import functools
import logging

import aiocells


logger = logging.getLogger()


async def async_printer(variable):
    print(f"value: {variable.value}")


def main():

    # Tests an async internal node in a flow graph

    graph = aiocells.DependencyGraph()

    time = aiocells.Variable()
    printer = graph.add_node(functools.partial(async_printer, time))

    graph.add_precedence(time, printer)

    # This example will continue until it is interrupted with Ctrl-C.
    #
    # Note that marking a function to be repeater function currently only
    # affects the behaviour of 'compute_flow' and not any of the other
    # `compute` functions.

    print()
    print("Ctrl-C to exit the demo")
    print()
    repeat_timer = aiocells.repeat(functools.partial(aiocells.timer, 1, time))
    graph.add_precedence(repeat_timer, time)

    asyncio.run(aiocells.compute_flow(graph))



#!/usr/bin/env python3

import asyncio
import functools
import logging

import aiocells


logger = logging.getLogger()


def main():

    clock = aiocells.ModClock()
    graph = aiocells.DependencyGraph()

    # Here, we simplify the previous demo by using a single variable to
    # store the time and a single printer to announce the modifications
    # when they happen. Because we are using 'compute_flow', the graph
    # is computed when any of the timers go off.
    time = aiocells.ModVariable(clock)
    printer = aiocells.ModPrinter(clock, time, "time changed to {value}")
    graph.add_precedence(time, printer)

    # Set the time after 0 seconds
    timer_0 = functools.partial(aiocells.timer, 0, time)
    graph.add_precedence(timer_0, time)

    # Set the time after 1 second
    timer_1 = functools.partial(aiocells.timer, 1, time)
    graph.add_precedence(timer_1, time)

    # Set the time after 3 seconds
    timer_3 = functools.partial(aiocells.timer, 3, time)
    graph.add_precedence(timer_3, time)

    asyncio.run(aiocells.compute_flow(graph))



#!/usr/bin/env python3

import aiocells

# Demonstrates modification tracking nodes. Nodes only compute if one or
# more of their dependencies have actually changed as signalled by the
# 'mod_time' attribute. To record 'mod_time' when a node changes, a mod.Clock
# is used, which always returns a new, higher value when the method 'now' is
# called.


def main():
    graph = aiocells.DependencyGraph()

    clock = aiocells.ModClock()

    variable_1 = aiocells.ModVariable(clock)
    variable_2 = aiocells.ModVariable(clock)

    printer_1 = aiocells.ModPrinter(clock, variable_1,
                                    "variable_1 changed to {value}")
    printer_2 = aiocells.ModPrinter(clock, variable_2,
                                    "variable_2 changed to {value}")

    graph.add_precedence(variable_1, printer_1)
    graph.add_precedence(variable_2, printer_2)

    print("Nothing has changed:")
    aiocells.compute_sequential(graph)

    variable_1.value = 1
    variable_2.value = 2
    print("Both variables:")
    aiocells.compute_sequential(graph)

    variable_1.value = 3
    print("variable_1 only:")
    aiocells.compute_sequential(graph)



#!/usr/bin/env python3

import asyncio
import functools
import logging

import aiocells


logger = logging.getLogger()


# See demo_18 for different way to do this
def main():

    clock = aiocells.ModClock()
    graph = aiocells.DependencyGraph()

    time = aiocells.ModVariable(clock)
    timer = functools.partial(aiocells.timer, 1, time)
    printer = aiocells.ModPrinter(clock, time, "time changed to {value}")

    graph.add_precedence(timer, time)
    graph.add_precedence(time, printer)

    # Demonstrating repeated computation of the same graph. However, 'flow
    # graphs', shown in the next demo, are better suited to this.
    for i in range(10):
        logger.info("Computation %s", i)
        asyncio.run(aiocells.async_compute_concurrent(graph))



#!/usr/bin/env python3

import aiocells


class HelloWorld:

    def __call__(self):
        print("Hello, world!")


def main():
    graph = aiocells.DependencyGraph()

    # In this example, we add a instance of a callable object rather than
    # a function
    node = graph.add_node(HelloWorld())

    aiocells.compute_sequential(graph)



import asyncio
import inspect
import logging

import aiocells.basic as basic
import aiocells.aio as aio


class ModClock:

    def __init__(self):
        self._now = 0

    def now(self):
        self._now += 1
        return self._now

    @property
    def current_value(self):
        return self._now


class ModVariable:

    def __init__(self, clock):
        self._clock = clock
        self._value = None
        self._new_value = None
        self.mod_time = None

    @property
    def value(self):
        return self._value

    @property
    def new_value(self):
        return self._new_value

    @value.setter
    def value(self, value):
        self._new_value = value

    @property
    def is_dirty(self):
        return self._value != self._new_value

    def __call__(self):
        if not self.is_dirty:
            return
        self.mod_time = self._clock.now()
        self._value = self._new_value


class ModAdder:

    def __init__(self, clock, input_cells, output_cell):
        self.clock = clock
        self.input_cells = input_cells
        self.output_cell = output_cell
        self.mod_time = None

    def __call__(self):
        requires_recalc = self.mod_time is None or any(
            input_cell.mod_time > self.mod_time
            for input_cell in self.input_cells
        )
        if not requires_recalc:
            return
        self.output_cell.value = sum(v.value for v in self.input_cells)
        self.mod_time = self.clock.now()


class ModPrinter:

    def __init__(self, clock, input_cell, message):
        self.clock = clock
        self.input_cell = input_cell
        self.message = message
        self.mod_time = None

    def __call__(self):
        if self.input_cell.mod_time is None:
            return
        assert self.input_cell.mod_time is not None
        if self.mod_time is not None \
           and self.mod_time >= self.input_cell.mod_time:
            return
        print(self.message.format(value=self.input_cell.value))
        self.mod_time = self.clock.now()



#!/usr/bin/env python3

import asyncio
import functools

import aiocells


def subgraph(name, period):

    clock = aiocells.ModClock()
    graph = aiocells.DependencyGraph()

    time = aiocells.ModVariable(clock)
    printer = aiocells.ModPrinter(
        clock, time, f"time in \"{name}\" changed to {{value}}"
    )
    graph.add_precedence(time, printer)

    timer_0 = functools.partial(aiocells.timer, 0, time)
    graph.add_precedence(timer_0, time)

    repeat_timer = aiocells.repeat(functools.partial(
        aiocells.timer, period, time
    ))
    graph.add_precedence(repeat_timer, time)

    return graph


def main():

    graph = aiocells.DependencyGraph()

    subgraph_1 = subgraph("graph_1", 0.7)
    subgraph_2 = subgraph("graph_2", 1.5)

    graph.add_node(functools.partial(aiocells.compute_flow, subgraph_1))
    graph.add_node(functools.partial(aiocells.compute_flow, subgraph_2))

    print()
    print("Ctrl-C to exit the demo")
    print()

    asyncio.run(aiocells.compute_flow(graph))



import click
import aiocells.demo_1
import aiocells.demo_2
import aiocells.demo_3
import aiocells.demo_4
import aiocells.demo_5
import aiocells.demo_6
import aiocells.demo_7
import aiocells.demo_8
import aiocells.demo_9
import aiocells.demo_10
import aiocells.demo_11
import aiocells.demo_12
import aiocells.demo_13
import aiocells.demo_14
import aiocells.demo_15
import aiocells.demo_16
import aiocells.demo_17
import aiocells.demo_18
import aiocells.demo_19
import aiocells.demo_20


@click.group()
def main():
    pass


@main.command()
def demo_1():
    aiocells.demo_1.main()


@main.command()
def demo_2():
    aiocells.demo_2.main()


@main.command()
def demo_3():
    aiocells.demo_3.main()


@main.command()
def demo_4():
    aiocells.demo_4.main()


@main.command()
def demo_5():
    aiocells.demo_5.main()


@main.command()
def demo_6():
    aiocells.demo_6.main()


@main.command()
def demo_7():
    aiocells.demo_7.main()


@main.command()
def demo_8():
    aiocells.demo_8.main()


@main.command()
def demo_9():
    aiocells.demo_9.main()


@main.command()
def demo_10():
    aiocells.demo_10.main()


@main.command()
def demo_11():
    aiocells.demo_11.main()


@main.command()
def demo_12():
    aiocells.demo_12.main()


@main.command()
def demo_13():
    aiocells.demo_13.main()


@main.command()
def demo_14():
    aiocells.demo_14.main()


@main.command()
def demo_15():
    aiocells.demo_15.main()


@main.command()
def demo_16():
    aiocells.demo_16.main()


@main.command()
def demo_17():
    aiocells.demo_17.main()


@main.command()
def demo_18():
    aiocells.demo_18.main()


@main.command()
def demo_19():
    aiocells.demo_19.main()


@main.command()
def demo_20():
    aiocells.demo_20.main()



import pathlib

from setuptools import setup, find_packages

THIS_DIR = pathlib.Path(__file__).parent
README = (THIS_DIR / "README.md").read_text()

setup(
    name="aiocells",
    version="0.1.20",
    python_requires=">=3.8",
    description="A package for synchronous and asynchronous"
                " dependency graph computation",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/isbjorntrading/aiocells",
    author="Anders Lindstrom",
    author_email="anders@isbjorn.com.au",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=[
        "click>=7.0",
    ],
    entry_points={
        "console_scripts": [
            "aiocells = aiocells.cli:main",
        ]
    },
)



import asyncio
import datetime
import inspect
import logging
import random

from .basic import TopologicalQueue


logger = logging.getLogger()


async def async_compute_sequential(graph):
    for node in graph.topological_ordering:
        assert callable(node) or inspect.iscoroutinefunction(node)
        if inspect.iscoroutinefunction(node):
            await node()
        else:
            assert callable(node)
            node()


def async_callable(the_callable):
    assert callable(the_callable)

    async def the_async_version():
        return the_callable()
    return the_async_version


def ensure_coroutine(node):
    assert callable(node) or inspect.iscoroutinefunction(node)
    if not inspect.iscoroutinefunction(node):
        node = async_callable(node)
    assert inspect.iscoroutinefunction(node)
    return node()


def create_tasks(nodes):
    return {asyncio.create_task(node) for node in nodes}


class MultipleTaskExceptions:

    def __init__(self, exceptions):
        self.exceptions = exceptions

    def __str__(self):
        return "MultiplTaskExceptions"


def raise_task_exceptions(tasks):
    all_exceptions = [task.exception() for task in tasks if
                      task.exception() is not None]
    if not all_exceptions:
        return
    raise Exception(f"There are {len(all_exceptions)} task exceptions.") \
        from all_exceptions[0]


async def async_compute_concurrent_simple(graph):

    assert graph is not None

    task_graph = graph.decorate(ensure_coroutine)

    queue = TopologicalQueue(task_graph)

    ready_tasks = create_tasks(queue.ready_set())

    while len(ready_tasks):
        assert not queue.empty()
        # 'pending' could be replaced with 'ready_tasks' here
        # and the line after (ready_tasks -= completed) could be removed
        completed, pending = await asyncio.wait(
            ready_tasks,
            return_when=asyncio.FIRST_COMPLETED
        )
        raise_task_exceptions(completed)
        ready_tasks -= completed
        for completed_task in completed:
            now_ready = queue.completed(completed_task.get_coro())
            # print(f"now_ready: {now_ready}")
            ready_tasks |= create_tasks(now_ready)
            # print(f"ready_tasks: {ready_tasks}")

    assert len(ready_tasks) == 0
    assert queue.empty(), f"queue.dependency_dict: {queue.dependency_dict}"


def prepare_ready_set(ready_set):
    callables = set()
    tasks = set()
    for f in ready_set:
        if inspect.iscoroutinefunction(f):
            # Generate a coroutine and task for each coro function
            task = asyncio.create_task(f())
            tasks.add(task)
            # Once the task is done, we need to mark the coro function
            # completed in the graph. To do that, we attach the coro function
            # to the task
            task.aio_coroutine_function = f
        elif callable(f):
            callables.add(f)
        else:
            raise ValueError(f"Node is neither a callable nor a"
                             " coro_function: {f}")
    return callables, tasks


async def async_compute_concurrent(graph):
    assert graph is not None

    queue = TopologicalQueue(graph)
    ready_set = queue.ready_set()

    logger.debug("ready_set: %s", ready_set)
    callables, running_tasks = prepare_ready_set(ready_set)

    while not queue.empty():

        while len(callables):
            the_callable = callables.pop()
            the_callable()
            now_ready = queue.completed(the_callable)
            new_callables, new_tasks = prepare_ready_set(now_ready)
            callables |= new_callables
            running_tasks |= new_tasks

        while len(running_tasks) and len(callables) == 0:
            completed_tasks, running_tasks = await asyncio.wait(
                running_tasks,
                return_when=asyncio.FIRST_COMPLETED
            )
            raise_task_exceptions(completed_tasks)
            for completed_task in completed_tasks:
                completed_coro_function = completed_task.aio_coroutine_function
                now_ready = queue.completed(completed_coro_function)
                new_callables, new_tasks = prepare_ready_set(now_ready)
                callables |= new_callables
                running_tasks |= new_tasks

    assert len(running_tasks) == 0
    assert queue.empty()


async def timer(seconds, result_variable):
    await asyncio.sleep(seconds)
    result_variable.value = datetime.datetime.now()


async def cancel_tasks(tasks):
    for task in tasks:
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass
